@@ -2,20 +2,21 @@
 
 // <groff_src_dir>/src/libs/libdriver/input.cpp
 
-/* Copyright (C) 1989, 1990, 1991, 1992, 2001, 2002, 2003
+/* Copyright (C) 1989, 1990, 1991, 1992, 2001, 2002, 2003, 2004, 2005,
+                 2006, 2008, 2009
    Free Software Foundation, Inc.
 
    Written by James Clark (jjc@jclark.com)
    Major rewrite 2001 by Bernd Warken (bwarken@mayn.de)
 
-   Last update: 04 Apr 2003
+   Last update: 5 Jan 2009
 
    This file is part of groff, the GNU roff text processing system.
 
    groff is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
+   the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
 
    groff is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -23,9 +24,7 @@
    General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with groff; see the file COPYING.  If not, write to the Free
-   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
 
 /* Description
@@ -165,7 +164,7 @@
 
   There was some confusion about the positioning of the graphical
   pointer at the printout after having executed a `D' command.
-  The classical troff manual of Osanna & Kernighan specified,
+  The classical troff manual of Ossanna & Kernighan specified,
 
     `The position after a graphical object has been drawn is
      at its end; for circles and ellipses, the "end" is at the
@@ -272,16 +271,15 @@ public:
   IntArray(void);
   IntArray(const size_t);
   ~IntArray(void);
-  const IntArg operator[](const size_t i) const
+  IntArg operator[](const size_t i) const
   {
     if (i >= num_stored)
       fatal("index out of range");
     return (IntArg) data[i];
   }
   void append(IntArg);
-  const IntArg * const
-    get_data(void) const { return (IntArg *) data; }
-  const size_t len(void) const { return num_stored; }
+  IntArg *get_data(void) const { return (IntArg *)data; }
+  size_t len(void) const { return num_stored; }
 };
 
 // Characters read from the input queue.
@@ -442,7 +440,7 @@ EnvStack::EnvStack(void)
 {
   num_allocated = 4;
   // allocate pointer to array of num_allocated pointers to environment
-  data = (environment **) malloc(envp_size * num_allocated);
+  data = (environment **)malloc(envp_size * num_allocated);
   if (data == 0)
     fatal("could not allocate environment data");
   num_stored = 0;
@@ -475,7 +473,7 @@ EnvStack::push(environment *e)
   if (num_stored >= num_allocated) {
     environment **old_data = data;
     num_allocated *= 2;
-    data = (environment **) malloc(envp_size * num_allocated);
+    data = (environment **)malloc(envp_size * num_allocated);
     if (data == 0)
       fatal("could not allocate data");
     for (size_t i = 0; i < num_stored; i++)
@@ -608,6 +606,7 @@ void delete_current_env(void)
   delete current_env->col;
   delete current_env->fill;
   delete current_env;
+  current_env = 0;
 }
 
 //////////////////////////////////////////////////////////////////////
@@ -1034,13 +1033,13 @@ remember_filename(const char *filename)
 {
   char *fname;
   if (strcmp(filename, "-") == 0)
-    fname = "<standard input>";
+    fname = (char *)"<standard input>";
   else
-    fname = (char *) filename;
+    fname = (char *)filename;
   size_t len = strlen(fname) + 1;
   if (current_filename != 0)
     free((char *)current_filename);
-  current_filename = (const char *) malloc(len);
+  current_filename = (const char *)malloc(len);
   if (current_filename == 0)
     fatal("can't malloc space for filename");
   strncpy((char *)current_filename, (char *)fname, len);
@@ -1060,13 +1059,13 @@ remember_source_filename(const char *filename)
 {
   char *fname;
   if (strcmp(filename, "-") == 0)
-    fname = "<standard input>";
+    fname = (char *)"<standard input>";
   else
-    fname = (char *) filename;
+    fname = (char *)filename;
   size_t len = strlen(fname) + 1;
   if (current_source_filename != 0)
     free((char *)current_source_filename);
-  current_source_filename = (const char *) malloc(len);
+  current_source_filename = (const char *)malloc(len);
   if (current_source_filename == 0)
     fatal("can't malloc space for filename");
   strncpy((char *)current_source_filename, (char *)fname, len);
@@ -1083,7 +1082,7 @@ void
 send_draw(const Char subcmd, const IntArray * const args)
 {
   EnvInt n = (EnvInt) args->len();
-  pr->draw((int) subcmd, (IntArg *) args->get_data(), n, current_env);
+  pr->draw((int) subcmd, (IntArg *)args->get_data(), n, current_env);
 }
 
 //////////////////////////////////////////////////////////////////////
@@ -1466,7 +1465,7 @@ parse_x_command(void)
     }
   case 'F':			// x Filename: set filename for errors
     {
-      char *str_arg = get_string_arg();
+      char *str_arg = get_extended_arg();
       if (str_arg == 0)
 	warning("empty argument for `x F' command");
       else {
@@ -1520,6 +1519,9 @@ parse_x_command(void)
       char *str_arg = get_extended_arg(); // includes line skip
       if (npages <= 0)
 	error("`x X' command invalid before first `p' command");
+      else if (str_arg && (strncmp(str_arg, "devtag:",
+				   strlen("devtag:")) == 0))
+	pr->devtag(str_arg, current_env);
       else
 	pr->special(str_arg, current_env);
       a_delete str_arg;
@@ -1730,7 +1732,7 @@ do_file(const char *filename)
       break;
     case 'F':			// F: obsolete, replaced by `x F'
       {
-	char *str_arg = get_string_arg();
+	char *str_arg = get_extended_arg();
 	remember_source_filename(str_arg);
 	a_delete str_arg;
 	break;
@@ -1823,6 +1825,7 @@ do_file(const char *filename)
   if (npages > 0)
     pr->end_page(current_env->vpos);
   delete pr;
+  pr = 0;
   fclose(current_file);
   // If `stopped' is not `true' here then there wasn't any `x stop'.
   if (!stopped)
